/*=============================================================================
Vitalnix User Management Suite
libvxmdsync/compare.c
  Copyright © Jan Engelhardt <jengelh [at] gmx de>, 2003 - 2006
  -- License restrictions apply (LGPL v2.1)

  This file is part of Vitalnix.
  Vitalnix is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  Vitalnix is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
  Boston, MA  02110-1301  USA

  -- For details, see the file named "LICENSE.LGPL2"
=============================================================================*/
#include <stdio.h>
#include <libHX.h>
#include "libvxmdsync/libvxmdsync.h"
#include "libvxpdb/libvxpdb.h"
#include "libvxpdb/xafunc.h"
#include "libvxutil/defines.h"
#include "libvxutil/libvxutil.h"

// Definitions
enum {
    ACTION_UPDATE       = 1 << 0, // do not re-add this user
    ACTION_DEFER_START  = 1 << 1, // start deferred deletion
    ACTION_DEFER_WAIT   = 1 << 2, // waits for deletion
    ACTION_DEFER_STOP   = 1 << 3, // cancel deferred deletion
    ACTION_DELETE_NOW   = 1 << 4, // select for final deletion
};

// Functions
static inline int decide_action(const struct vxpdb_user *,
    const struct HXbtree *, long);

//-----------------------------------------------------------------------------
void mdsync_compare(struct mdsync_workspace *w) {
    const struct vxpdb_group *grp = &w->dest_group;
    const struct HXbtree_node *ln_node;
    long users_proc, users_max;
    int analyze, decision;
    struct vxpdb_user pwd = {};
    void *travp;

    users_proc = 0;
    users_max  = vxpdb_modctl(w->database, PDB_COUNT_USERS);
    travp      = vxpdb_usertrav_init(w->database);

    while(vxpdb_usertrav_walk(w->database, travp, &pwd)) {
        // Record username in @lnlist
        ln_node = HXbtree_add(w->lnlist, pwd.pw_name);
        analyze = (grp->gr_gid != PDB_NOGID && grp->gr_gid == pwd.pw_gid) ||
                  (grp->gr_name != NULL && pwd.pw_igrp != NULL &&
                   strcmp(grp->gr_name, pwd.pw_igrp) == 0);

        if(w->report != NULL)
            w->report(MDREP_COMPARE, w, ++users_proc, users_max);
        if(!analyze)
            continue;

        // Users we look for (only specified group)
        decision = decide_action(&pwd, w->add_req,
                   w->config.add_opts.defaults.vs_defer);
        ++w->num_grp;

        if(decision & ACTION_UPDATE) {
            HXbtree_add(w->keep_req, pwd.vs_uuid, vxpdb_user_dup(&pwd));
            HXbtree_del(w->add_req, pwd.vs_uuid);
        }
        if(decision & ACTION_DEFER_START)
            HXdeque_push(w->defer_start, ln_node->key);
        if(decision & ACTION_DEFER_WAIT)
            HXdeque_push(w->defer_wait, ln_node->key);
        if(decision & ACTION_DEFER_STOP)
            HXdeque_push(w->defer_stop, ln_node->key);
        if(decision & ACTION_DELETE_NOW)
            HXdeque_push(w->delete_now, ln_node->key);

        /* username is shared between all these data structures, but to define
        a common point, they are owned by @w->lnlist. (Because this is the
        only structure which keeps _all_ usernames.) */
    }

    vxpdb_usertrav_free(w->database, travp);
    vxpdb_user_free(&pwd, 0);
    if(w->report != NULL)
        w->report(MDREP_COMPARE, w, users_max, users_max);
    return;
}

void mdsync_compare_simple(struct mdsync_workspace *w) {
    void *travp = vxpdb_usertrav_init(w->database);
    struct vxpdb_user pwd = {};
    int count, count_max;

    count     = 0;
    count_max = vxpdb_modctl(w->database, PDB_COUNT_USERS);

    if(travp != NULL)
        while(vxpdb_usertrav_walk(w->database, travp, &pwd)) {
            // No invalid references later, thanks to HXBT_CDATA
            HXbtree_add(w->lnlist, pwd.pw_name);
            if(w->report != NULL)
                w->report(MDREP_COMPARE, w, ++count, count_max);
        }

    vxpdb_usertrav_free(w->database, travp);
    vxpdb_user_free(&pwd, 0);
    return;
}

//-----------------------------------------------------------------------------
static inline int decide_action(const struct vxpdb_user *user,
 const struct HXbtree *eds_data, long defer)
{
    if(user->vs_uuid != NULL && HXbtree_find(eds_data, user->vs_uuid) != NULL)
        // PDB user found in EDS: Keep, and remove from add_req.
        return ACTION_UPDATE | ((user->vs_defer > 0) ? ACTION_DEFER_STOP : 0);

    // PDB user not found in EDS
    if(defer > 0) {
        if(user->vs_defer == 0)
            return ACTION_DEFER_START;
        else if(now_to_iday() >= user->vs_defer + defer)
            // Did not reappear on EDS list within timelimit
            return ACTION_DELETE_NOW;
        else
            return ACTION_DEFER_WAIT;
    } else {
        return ACTION_DELETE_NOW;
    }

    return -1;
}

//=============================================================================
